# This code is part of the Biopython distribution and governed by its
# license.  Please see the LICENSE file that should have been included
# as part of this package.
#

"""GenomeDiagram script to mimic Proux et al 2002 Figure 6.

You can use the Entrez module to download the 3 required GenBank files

This is an extended version of the example in the Biopython Tutorial
which produces a GenomeDiagram figure close to Proux et al 2002 Figure 6.

See https://doi.org/10.1128/JB.184.21.6026-6036.2002
"""

from reportlab.lib import colors
from reportlab.lib.colors import blue
from reportlab.lib.colors import brown
from reportlab.lib.colors import green
from reportlab.lib.colors import grey
from reportlab.lib.colors import lightblue
from reportlab.lib.colors import orange
from reportlab.lib.colors import purple
from reportlab.lib.colors import red

from Bio import SeqIO
from Bio.Graphics import GenomeDiagram
from Bio.Graphics.GenomeDiagram import CrossLink
from Bio.SeqFeature import SeqFeature
from Bio.SeqFeature import SimpleLocation

name = "Proux Fig 6"

# As explained in the Biopython Tutorial, these are three phage genomes. The first
# two are self-containged GenBank files, but the third phage is integrated into a
# bacterial genome, thus we slice the full record and also take the reverse
# complement to match the strand orientation of the other two phage:

A_rec = SeqIO.read("NC_002703.gbk", "gb")
B_rec = SeqIO.read("AF323668.gbk", "gb")
C_rec = SeqIO.read("NC_003212.gbk", "gb")[2587879:2625807].reverse_complement(name=True)
records = {rec.name: rec for rec in [A_rec, B_rec, C_rec]}

# Here we hard code the gene colors for simiplicity and to match the target image.
# In practice you might have an automatic mapping based on the gene annotation
# or some other classification:

A_colors = (
    [red] * 5
    + [grey] * 7
    + [orange] * 2
    + [grey] * 2
    + [orange]
    + [grey] * 11
    + [green] * 4
    + [grey]
    + [green] * 2
    + [grey, green]
    + [brown] * 5
    + [blue] * 4
    + [lightblue] * 5
    + [grey, lightblue]
    + [purple] * 2
    + [grey]
)
B_colors = (
    [red] * 6
    + [grey] * 8
    + [orange] * 2
    + [grey]
    + [orange]
    + [grey] * 21
    + [green] * 5
    + [grey]
    + [brown] * 4
    + [blue] * 3
    + [lightblue] * 3
    + [grey] * 5
    + [purple] * 2
)
C_colors = (
    [grey] * 33
    + [green] * 5
    + [brown] * 4
    + [blue] * 2
    + [grey, blue]
    + [lightblue] * 2
    + [grey] * 8
)

# Here we hard code a list of cross-links with percentage identity scores, based
# on a manual inspection of the target image (there could be mistakes here).
# In practice you might generate the list of cross-mappings from a BLAST report
# or similar computational analysis:

# Tuc2009 (NC_002703) vs bIL285 (AF323668)
A_vs_B = [
    (99, "Tuc2009_01", "int"),
    (33, "Tuc2009_03", "orf4"),
    (94, "Tuc2009_05", "orf6"),
    (100, "Tuc2009_06", "orf7"),
    (97, "Tuc2009_07", "orf8"),
    (98, "Tuc2009_08", "orf9"),
    (98, "Tuc2009_09", "orf10"),
    (100, "Tuc2009_10", "orf12"),
    (100, "Tuc2009_11", "orf13"),
    (94, "Tuc2009_12", "orf14"),
    (87, "Tuc2009_13", "orf15"),
    (94, "Tuc2009_14", "orf16"),
    (94, "Tuc2009_15", "orf17"),
    (88, "Tuc2009_17", "rusA"),
    (91, "Tuc2009_18", "orf20"),
    (93, "Tuc2009_19", "orf22"),
    (71, "Tuc2009_20", "orf23"),
    (51, "Tuc2009_22", "orf27"),
    (97, "Tuc2009_23", "orf28"),
    (88, "Tuc2009_24", "orf29"),
    (26, "Tuc2009_26", "orf38"),
    (19, "Tuc2009_46", "orf52"),
    (77, "Tuc2009_48", "orf54"),
    (91, "Tuc2009_49", "orf55"),
    (95, "Tuc2009_52", "orf60"),
]

# bIL285 (AF323668) vs Listeria innocua prophage 5 (in NC_003212)
B_vs_C = [
    (42, "orf39", "lin2581"),
    (31, "orf40", "lin2580"),
    (49, "orf41", "lin2579"),  # terL
    (54, "orf42", "lin2578"),  # portal
    (55, "orf43", "lin2577"),  # protease
    (33, "orf44", "lin2576"),  # mhp
    (51, "orf46", "lin2575"),
    (33, "orf47", "lin2574"),
    (40, "orf48", "lin2573"),
    (25, "orf49", "lin2572"),
    (50, "orf50", "lin2571"),
    (48, "orf51", "lin2570"),
    (24, "orf52", "lin2568"),
    (30, "orf53", "lin2567"),
    (28, "orf54", "lin2566"),
]


def get_feature(features, id, tags=("locus_tag", "gene", "old_locus_tag")):
    """Search list of SeqFeature objects for an identifier under the given tags."""
    for f in features:
        for key in tags:
            # tag may not be present in this feature
            for x in f.qualifiers.get(key, []):
                if x == id:
                    return f
    raise KeyError(id)


gd_diagram = GenomeDiagram.Diagram(name)
feature_sets = {}
max_len = 0
for i, record in enumerate([A_rec, B_rec, C_rec]):
    max_len = max(max_len, len(record))
    # Allocate tracks 5 (top), 3, 1 (bottom) for A, B, C
    # (empty tracks 2 and 4 add useful white space to emphasise the cross links
    # and also serve to make the tracks vertically more compressed)
    gd_track_for_features = gd_diagram.new_track(
        5 - 2 * i,
        name=record.name,
        greytrack=True,
        height=0.5,
        start=0,
        end=len(record),
    )
    assert record.name not in feature_sets
    feature_sets[record.name] = gd_track_for_features.new_set()

# We add dummy features to the tracks for each cross-link BEFORE we add the
# arrow features for the genes. This ensures the genes appear on top:
for X, Y, X_vs_Y in [
    ("NC_002703", "AF323668", A_vs_B),
    ("AF323668", "NC_003212", B_vs_C),
]:
    features_X = records[X].features
    features_Y = records[Y].features
    set_X = feature_sets[X]
    set_Y = feature_sets[Y]
    for score, x, y in X_vs_Y:
        color = colors.linearlyInterpolatedColor(
            colors.white, colors.firebrick, 0, 100, score
        )
        border = colors.lightgrey
        f_x = get_feature(features_X, x)
        F_x = set_X.add_feature(
            SeqFeature(SimpleLocation(f_x.location.start, f_x.location.end, strand=0)),
            color=color,
            border=border,
        )
        f_y = get_feature(features_Y, y)
        F_y = set_Y.add_feature(
            SeqFeature(SimpleLocation(f_y.location.start, f_y.location.end, strand=0)),
            color=color,
            border=border,
        )
        gd_diagram.cross_track_links.append(CrossLink(F_x, F_y, color, border))


for record, gene_colors in zip([A_rec, B_rec, C_rec], [A_colors, B_colors, C_colors]):
    gd_feature_set = feature_sets[record.name]

    i = 0
    for feature in record.features:
        if feature.type != "gene":
            # Exclude this feature
            continue
        try:
            g_color = gene_colors[i]
        except IndexError:
            print("Don't have color for %s gene %i" % (record.name, i))
            g_color = grey
        gd_feature_set.add_feature(
            feature,
            sigil="BIGARROW",
            color=g_color,
            label=True,
            name=str(i + 1),
            label_position="start",
            label_size=6,
            label_angle=0,
        )
        i += 1

gd_diagram.draw(format="linear", pagesize="A4", fragments=1, start=0, end=max_len)
gd_diagram.write(name + ".pdf", "PDF")
gd_diagram.write(name + ".eps", "EPS")
gd_diagram.write(name + ".svg", "SVG")



# Copyright 2008 by Bartek Wilczynski.
# Revisions copyright 2019 by Victor Lin.
# Adapted from  Bio.MEME.Parser by Jason A. Hackney.  All rights reserved.
# This code is part of the Biopython distribution and governed by its
# license.  Please see the LICENSE file that should have been included
# as part of this package.
"""Module for the support of MEME motif format."""

import xml.etree.ElementTree as ET

from Bio import Align
from Bio import motifs
from Bio.Seq import Seq


def read(handle):
    """Parse the text output of the MEME program into a meme.Record object.

    Examples
    --------
    >>> from Bio.motifs import meme
    >>> with open("motifs/meme.INO_up800.classic.oops.xml") as f:
    ...     record = meme.read(f)
    >>> for motif in record:
    ...     for sequence in motif.alignment.sequences:
    ...         print(sequence.motif_name, sequence.sequence_name, sequence.sequence_id, sequence.strand, sequence.pvalue)
    GSKGCATGTGAAA INO1 sequence_5 + 1.21e-08
    GSKGCATGTGAAA FAS1 sequence_2 - 1.87e-08
    GSKGCATGTGAAA ACC1 sequence_4 - 6.62e-08
    GSKGCATGTGAAA CHO2 sequence_1 - 1.05e-07
    GSKGCATGTGAAA CHO1 sequence_0 - 1.69e-07
    GSKGCATGTGAAA FAS2 sequence_3 - 5.62e-07
    GSKGCATGTGAAA OPI3 sequence_6 + 1.08e-06
    TTGACWCYTGCYCWG CHO2 sequence_1 + 7.2e-10
    TTGACWCYTGCYCWG OPI3 sequence_6 - 2.56e-08
    TTGACWCYTGCYCWG ACC1 sequence_4 - 1.59e-07
    TTGACWCYTGCYCWG CHO1 sequence_0 + 2.05e-07
    TTGACWCYTGCYCWG FAS1 sequence_2 + 3.85e-07
    TTGACWCYTGCYCWG FAS2 sequence_3 - 5.11e-07
    TTGACWCYTGCYCWG INO1 sequence_5 + 8.01e-07

    """
    record = Record()
    try:
        xml_tree = ET.parse(handle)
    except ET.ParseError:
        raise ValueError(
            "Improper MEME XML input file. XML root tag should start with <MEME version= ..."
        )
    __read_metadata(record, xml_tree)
    __read_alphabet(record, xml_tree)
    sequence_id_name_map = __get_sequence_id_name_map(xml_tree)
    record.sequences = list(sequence_id_name_map.keys())
    __read_motifs(record, xml_tree, sequence_id_name_map)
    return record


class Motif(motifs.Motif):
    """A subclass of Motif used in parsing MEME (and MAST) output.

    This subclass defines functions and data specific to MEME motifs.
    This includes the motif name, the evalue for a motif, and its number
    of occurrences.
    """

    def __init__(self, alphabet=None, alignment=None):
        """Initialize the class."""
        motifs.Motif.__init__(self, alphabet, alignment)
        self.evalue = 0.0
        self.num_occurrences = 0
        self.name = None
        self.id = None
        self.alt_id = None


class Record(list):
    """A class for holding the results of a MEME run.

    A meme.Record is an object that holds the results from running
    MEME. It implements no methods of its own.

    The meme.Record class inherits from list, so you can access individual
    motifs in the record by their index. Alternatively, you can find a motif
    by its name:

    >>> from Bio import motifs
    >>> with open("motifs/meme.INO_up800.classic.oops.xml") as f:
    ...     record = motifs.parse(f, 'MEME')
    >>> motif = record[0]
    >>> print(motif.name)
    GSKGCATGTGAAA
    >>> motif = record['GSKGCATGTGAAA']
    >>> print(motif.name)
    GSKGCATGTGAAA
    """

    def __init__(self):
        """Initialize the class."""
        self.version = ""
        self.datafile = ""
        self.command = ""
        self.alphabet = ""
        self.sequences = []

    def __getitem__(self, key):
        """Return the motif of index key."""
        if isinstance(key, str):
            for motif in self:
                if motif.name == key:
                    return motif
        else:
            return list.__getitem__(self, key)


# Everything below is private


def __read_metadata(record, xml_tree):
    record.version = xml_tree.getroot().get("version")
    record.datafile = xml_tree.find("training_set").get("primary_sequences")
    record.command = xml_tree.find("model").find("command_line").text
    # TODO - background_frequencies, other metadata under model


def __read_alphabet(record, xml_tree):
    alphabet_tree = (
        xml_tree.find("training_set").find("letter_frequencies").find("alphabet_array")
    )
    for value in alphabet_tree.findall("value"):
        record.alphabet += value.get("letter_id")


def __get_sequence_id_name_map(xml_tree):
    return {
        sequence_tree.get("id"): sequence_tree.get("name")
        for sequence_tree in xml_tree.find("training_set").findall("sequence")
    }


def __read_motifs(record, xml_tree, sequence_id_name_map):
    for motif_tree in xml_tree.find("motifs").findall("motif"):
        instances = []
        for site_tree in motif_tree.find("contributing_sites").findall(
            "contributing_site"
        ):
            letters = [
                letter_ref.get("letter_id")
                for letter_ref in site_tree.find("site").findall("letter_ref")
            ]
            sequence = "".join(letters)
            instance = Seq(sequence)
            instance.motif_name = motif_tree.get("name")
            instance.sequence_id = site_tree.get("sequence_id")
            instance.sequence_name = sequence_id_name_map[instance.sequence_id]
            # TODO - left flank, right flank
            instance.start = int(site_tree.get("position")) + 1
            instance.pvalue = float(site_tree.get("pvalue"))
            instance.strand = __convert_strand(site_tree.get("strand"))
            instance.length = len(sequence)
            instances.append(instance)
        alignment = Align.Alignment(instances)
        motif = Motif(record.alphabet, alignment)
        motif.id = motif_tree.get("id")
        motif.name = motif_tree.get("name")
        motif.alt_id = motif_tree.get("alt")
        motif.length = int(motif_tree.get("width"))
        motif.num_occurrences = int(motif_tree.get("sites"))
        motif.evalue = float(motif_tree.get("e_value"))
        # TODO - ic, re, llr, pvalue, bayes_threshold, elapsed_time
        record.append(motif)


def __convert_strand(strand):
    """Convert strand (+/-) from XML if present.

    Default: +
    """
    if strand == "minus":
        return "-"
    if strand == "plus" or strand == "none":
        return "+"


if __name__ == "__main__":
    from Bio._utils import run_doctest

    run_doctest()



import codecs
import html
from abc import ABC
from abc import abstractmethod

from Bio.Seq import UndefinedSequenceError


def _html_entity_replace(error):
    start = error.start
    end = error.end
    character = error.object[start:end]
    codepoint = ord(character)
    entity = f"&amp;{html.entities.codepoint2name[codepoint]};"
    return entity, end


codecs.register_error("htmlentityreplace", _html_entity_replace)


class BaseXMLWriter(ABC):
    """XML Writer, abstract base class."""

    def __init__(self, stream):
        """Init the stuff."""
        self.stream = stream

    def write(self, records):
        """Write the records."""
        stream = self.stream
        program = records.program
        self._program = program
        self._write_xml_declaration()
        self._write_definition()
        self._start_blastoutput()
        self._write_program(program.encode())
        self._write_version(records.version.encode())
        reference = html.escape(records.reference).encode("ASCII", "htmlentityreplace")
        self._write_reference(reference)
        self._write_db(records.db.encode())
        try:
            query = records.query
        except AttributeError:  # XML2
            pass
        else:
            self._write_query_id(query.id.encode())
            self._write_query_def(query.description.encode())
            self._write_query_len(len(query))
            try:
                query_seq = bytes(query.seq)
            except UndefinedSequenceError:
                pass
            else:
                self._write_query_seq(query_seq)
        self._write_params(records.param)
        count = self._write_records(records)
        try:
            mbstat = records.mbstat
        except AttributeError:
            pass
        else:
            self._start_mbstat()
            self._write_statistics(mbstat)
            self._end_mbstat()
        self._end_blastoutput()
        return count

    def _write_xml_declaration(self):
        self.stream.write(b'<?xml version="1.0"?>\n')

    def _write_params(self, param):
        self._start_param()
        value = param.get("matrix")
        if value is not None:
            self._write_parameters_matrix(value.encode())
        self._write_parameters_expect(param["expect"])
        value = param.get("include")
        if value is not None:
            self._write_parameters_include(value)
        value = param.get("sc-match")
        if value is not None:
            self._write_parameters_sc_match(value)
        value = param.get("sc-mismatch")
        if value is not None:
            self._write_parameters_sc_mismatch(value)
        value = param.get("gap-open")
        if value is not None:
            self._write_parameters_gap_open(value)
        value = param.get("gap-extend")
        if value is not None:
            self._write_parameters_gap_extend(value)
        value = param.get("filter")
        if value is not None:
            self._write_parameters_filter(value.encode())
        value = param.get("pattern")
        if value is not None:
            self._write_parameters_pattern(value.encode())
        value = param.get("entrez-query")
        if value is not None:
            self._write_parameters_entrez_query(value.encode())
        value = param.get("cbs")
        if value is not None:
            self._write_parameters_cbs(value)
        value = param.get("query-gencode")
        if value is not None:
            self._write_parameters_query_gencode(value)
        value = param.get("db-gencode")
        if value is not None:
            self._write_parameters_db_gencode(value)
        value = param.get("bl2seq-mode")
        if value is not None:
            self._write_parameters_bl2seq_mode(value.encode())
        self._end_param()

    def _write_records(self, records):
        count = 0
        self._start_iterations()
        for record in records:
            self._write_record(record)
            count += 1
        self._end_iterations()
        return count

    def _write_record(self, record):
        stream = self.stream
        self._start_iteration(record)
        query = record.query
        if query is None:
            query_length = None
        else:
            query_length = len(query.seq)
            self._write_iteration_query_id(query.id.encode())
            description = query.description
            if description != "<unknown description>":
                self._write_iteration_query_def(description.encode())
            self._write_iteration_query_len(query_length)
            for feature in query.features:
                if feature.type == "masking":
                    self._write_iteration_query_masking(feature.location)
        self._start_iteration_hits()
        for hit in record:
            self._write_hit(hit, query_length)
        self._end_iteration_hits()
        try:
            stat = record.stat
        except AttributeError:
            pass
        else:
            self._start_iteration_stat()
            self._write_statistics(stat)
            self._end_iteration_stat()
        self._end_iteration()

    def _write_hit(self, hit, query_length):
        stream = self.stream
        program = self._program
        target = hit.target
        target_length = len(target.seq)
        self._start_iteration_hit()
        self._write_hit_num(hit.num)
        try:
            targets = hit.targets
        except AttributeError:  # XML
            self._write_hit_id(target.id.encode())
            self._write_hit_def(target.description.encode())
            self._write_hit_accession(target.name.encode())
        else:  # XML2
            self._start_hit_targets()
            for target in targets:
                self._start_hitdescr()
                self._write_hit_id(target.id.encode())
                self._write_hit_accession(target.name.encode())
                self._write_hit_def(target.description.encode())
                taxid = target.annotations.get("taxid")
                if taxid is not None:
                    self._write_hit_taxid(taxid)
                sciname = target.annotations.get("sciname")
                if sciname is not None:
                    self._write_hit_sciname(sciname.encode())
                self._end_hitdescr()
            self._end_hit_targets()
        self._write_hit_len(target_length)
        self._start_hit_hsps()
        for hsp in hit:
            self._write_hsp(hsp, query_length, target_length)
        self._end_hit_hsps()
        self._end_iteration_hit()

    def _write_hsp(self, hsp, query_length, target_length):
        stream = self.stream
        program = self._program
        query = hsp.query
        target = hsp.target
        coordinates = hsp.coordinates
        if coordinates.shape[1] > 0:
            hit_from, query_from = coordinates[:, 0]
            hit_to, query_to = coordinates[:, -1]
            if program in ("blastn", "megablast"):
                if hit_from <= hit_to:
                    hit_frame = 1
                    hit_from += 1
                else:
                    hit_frame = -1
                    hit_to += 1
            elif program in ("blastp", "blastx", "rpsblast", "psiblast"):
                hit_from += 1
                hit_frame = 0
            elif program in ("tblastn", "tblastx"):
                feature = target.features[0]
                coded_by = feature.qualifiers["coded_by"]
                if coded_by.startswith("complement("):
                    assert coded_by.endswith(")")
                    coded_by = coded_by[11:-1]
                    strand = -1
                else:
                    strand = +1
                hit_id, hit_from_to = coded_by.split(":")
                hit_from, hit_to = hit_from_to.split("..")
                hit_from = int(hit_from)
                hit_to = int(hit_to)
                hit_start = hit_from - 1
                hit_end = hit_to
                if strand == +1:
                    hit_frame = hit_start % 3 + 1
                else:
                    hit_frame = (hit_end - target_length) % -3 - 1
            if program in ("blastn", "megablast"):
                if query_from <= query_to:
                    query_from += 1
                    query_frame = 1
                else:
                    query_to += 1
                    query_frame = -1
            elif program in ("blastp", "tblastn", "rpsblast", "psiblast"):
                query_from += 1
                query_frame = 0
            elif program in ("blastx", "tblastx"):
                feature = query.features[0]
                coded_by = feature.qualifiers["coded_by"]
                if coded_by.startswith("complement("):
                    assert coded_by.endswith(")")
                    coded_by = coded_by[11:-1]
                    strand = -1
                else:
                    strand = +1
                query_id, query_from_to = coded_by.split(":")
                query_from, query_to = query_from_to.split("..")
                query_from = int(query_from)
                query_to = int(query_to)
                query_start = query_from - 1
                query_end = query_to
                if strand == +1:
                    query_frame = query_start % 3 + 1
                else:
                    query_frame = (query_end - query_length) % -3 - 1
            hseq = hsp[0]
            qseq = hsp[1]
            align_len = len(hseq)
        else:
            # PSIBLAST XML2
            query_from = 0
            query_to = 0
            hit_from = 0
            hit_to = 0
            hseq = ""
            qseq = ""
            query_frame = None
            hit_frame = None
            align_len = None
        annotations = hsp.annotations
        bit_score = annotations["bit score"]
        evalue = annotations["evalue"]
        identity = annotations["identity"]
        positive = annotations.get("positive")
        gaps = annotations.get("gaps")
        midline = annotations.get("midline")
        self._start_hsp()
        self._write_hsp_num(hsp.num)
        self._write_hsp_bit_score(str(bit_score).encode())
        self._write_hsp_score(hsp.score)
        self._write_hsp_evalue(str(evalue).encode())
        self._write_hsp_query_from(query_from)
        self._write_hsp_query_to(query_to)
        self._write_hsp_hit_from(hit_from)
        self._write_hsp_hit_to(hit_to)
        if query_frame is not None:
            self._write_hsp_query_frame(query_frame)
        if hit_frame is not None:
            self._write_hsp_hit_frame(hit_frame)
        self._write_hsp_identity(identity)
        if positive is not None:
            self._write_hsp_positive(positive)
        if gaps is not None:
            self._write_hsp_gaps(gaps)
        if align_len is not None:
            self._write_hsp_align_len(align_len)
        self._write_hsp_qseq(qseq.encode())
        self._write_hsp_hseq(hseq.encode())
        if midline is not None:
            self._write_hsp_midline(midline.encode())
        self._end_hsp()

    def _write_statistics(self, stat):
        self.stream.write(b"    <Statistics>\n")
        self._write_statistics_db_num(stat["db-num"])
        self._write_statistics_db_len(stat["db-len"])
        self._write_statistics_hsp_len(stat["hsp-len"])
        self._write_statistics_eff_space(str(stat["eff-space"]).encode())
        self._write_statistics_kappa(stat["kappa"])
        self._write_statistics_lambda(stat["lambda"])
        self._write_statistics_entropy(stat["entropy"])
        self.stream.write(b"    </Statistics>\n")

    def _write_definition(self):
        return

    @abstractmethod
    def _start_blastoutput(self):
        return

    @abstractmethod
    def _end_blastoutput(self):
        return

    @abstractmethod
    def _write_program(self, program):
        return

    @abstractmethod
    def _write_version(self, version):
        return

    @abstractmethod
    def _write_reference(self, reference):
        return

    @abstractmethod
    def _write_db(self, db):
        return

    @abstractmethod
    def _start_param(self):
        return

    @abstractmethod
    def _end_param(self):
        return

    @abstractmethod
    def _start_iterations(self):
        return

    @abstractmethod
    def _end_iterations(self):
        return

    @abstractmethod
    def _start_iteration(self):
        return

    @abstractmethod
    def _end_iteration(self):
        return

    @abstractmethod
    def _write_iteration_query_id(self, query_id):
        return

    @abstractmethod
    def _write_iteration_query_def(self, query_def):
        return

    @abstractmethod
    def _write_iteration_query_len(self, query_len):
        return

    def _write_iteration_query_masking(self, location):
        return

    @abstractmethod
    def _start_iteration_hits(self):
        return

    @abstractmethod
    def _end_iteration_hits(self):
        return

    @abstractmethod
    def _start_iteration_stat(self):
        return

    @abstractmethod
    def _end_iteration_stat(self):
        return

    @abstractmethod
    def _write_parameters_matrix(self, value):
        return

    @abstractmethod
    def _write_parameters_expect(self, value):
        return

    @abstractmethod
    def _write_parameters_include(self, value):
        return

    @abstractmethod
    def _write_parameters_sc_match(self, value):
        return

    @abstractmethod
    def _write_parameters_sc_mismatch(self, value):
        return

    @abstractmethod
    def _write_parameters_gap_open(self, value):
        return

    @abstractmethod
    def _write_parameters_gap_extend(self, value):
        return

    @abstractmethod
    def _write_parameters_filter(self, value):
        return

    @abstractmethod
    def _write_parameters_pattern(self, value):
        return

    @abstractmethod
    def _write_parameters_entrez_query(self, value):
        return

    def _write_parameters_cbs(self, value):
        return

    def _write_parameters_query_gencode(self, value):
        return

    def _write_parameters_db_gencode(self, value):
        return

    def _write_parameters_bl2seq_mode(self, value):
        return

    @abstractmethod
    def _write_statistics_db_num(self, db_num):
        return

    @abstractmethod
    def _write_statistics_db_len(self, db_len):
        return

    @abstractmethod
    def _write_statistics_hsp_len(self, hsp_len):
        return

    @abstractmethod
    def _write_statistics_eff_space(self, value):
        return

    @abstractmethod
    def _write_statistics_kappa(self, value):
        return

    @abstractmethod
    def _write_statistics_lambda(self, value):
        return

    @abstractmethod
    def _write_statistics_entropy(self, value):
        return

    @abstractmethod
    def _start_iteration_hit(self):
        return

    @abstractmethod
    def _end_iteration_hit(self):
        return

    @abstractmethod
    def _write_hit_num(self, num):
        return

    @abstractmethod
    def _write_hit_id(self, hit_id):
        return

    @abstractmethod
    def _write_hit_def(self, hit_def):
        return

    @abstractmethod
    def _write_hit_accession(self, hit_accession):
        return

    def _write_hit_taxid(self, taxid):
        return

    def _write_hit_sciname(self, sciname):
        return

    @abstractmethod
    def _write_hit_len(self, hit_length):
        return

    @abstractmethod
    def _start_hit_hsps(self):
        return

    @abstractmethod
    def _end_hit_hsps(self):
        return

    @abstractmethod
    def _start_hsp(self):
        return

    @abstractmethod
    def _end_hsp(self):
        return

    @abstractmethod
    def _write_hsp_num(self, num):
        return

    @abstractmethod
    def _write_hsp_bit_score(self, bit_score):
        return

    @abstractmethod
    def _write_hsp_score(self, score):
        return

    @abstractmethod
    def _write_hsp_evalue(self, evalue):
        return

    @abstractmethod
    def _write_hsp_query_from(self, query_from):
        return

    @abstractmethod
    def _write_hsp_query_to(self, query_to):
        return

    @abstractmethod
    def _write_hsp_hit_from(self, hit_from):
        return

    @abstractmethod
    def _write_hsp_hit_to(self, hit_to):
        return

    @abstractmethod
    def _write_hsp_pattern_from(self, hit_from):
        return

    @abstractmethod
    def _write_hsp_pattern_to(self, hit_to):
        return

    @abstractmethod
    def _write_hsp_query_frame(self, query_frame):
        return

    @abstractmethod
    def _write_hsp_hit_frame(self, hit_frame):
        return

    @abstractmethod
    def _write_hsp_identity(self, identity):
        return

    @abstractmethod
    def _write_hsp_positive(self, positive):
        return

    @abstractmethod
    def _write_hsp_gaps(self, gaps):
        return

    @abstractmethod
    def _write_hsp_align_len(self, align_len):
        return

    @abstractmethod
    def _write_hsp_density(self, density):
        return

    @abstractmethod
    def _write_hsp_qseq(self, qseq):
        return

    @abstractmethod
    def _write_hsp_hseq(self, hseq):
        return

    @abstractmethod
    def _write_hsp_midline(self, midline):
        return


class XMLWriter(BaseXMLWriter):
    """XML Writer."""

    def _write_definition(self):
        self.stream.write(
            b"""\
<!DOCTYPE BlastOutput PUBLIC "-//NCBI//NCBI BlastOutput/EN" "https://www.ncbi.nlm.nih.gov/dtd/NCBI_BlastOutput.dtd">
"""
        )

    def _start_blastoutput(self):
        self.stream.write(b"<BlastOutput>\n")

    def _end_blastoutput(self):
        self.stream.write(b"</BlastOutput>\n")

    def _write_program(self, program):
        self.stream.write(b"<BlastOutput_program>%b</BlastOutput_program>\n" % program)

    def _write_version(self, version):
        self.stream.write(b"<BlastOutput_version>%b</BlastOutput_version>\n" % version)

    def _write_reference(self, reference):
        self.stream.write(
            b"<BlastOutput_reference>%b</BlastOutput_reference>\n" % reference
        )

    def _write_db(self, db):
        self.stream.write(b"<BlastOutput_db>%b</BlastOutput_db>\n" % db)

    def _write_query_id(self, query_id):
        self.stream.write(
            b"<BlastOutput_query-ID>%b</BlastOutput_query-ID>\n" % query_id
        )

    def _write_query_def(self, query_def):
        self.stream.write(
            b"<BlastOutput_query-def>%b</BlastOutput_query-def>\n" % query_def
        )

    def _write_query_len(self, query_len):
        self.stream.write(
            b"<BlastOutput_query-len>%d</BlastOutput_query-len>\n" % query_len
        )

    def _write_query_seq(self, query_seq):
        self.stream.write(
            b"<BlastOutput_query-seq>%b</BlastOutput_query-seq>\n" % query_seq
        )

    def _start_param(self):
        self.stream.write(
            b"""\
  <BlastOutput_param>
    <Parameters>
"""
        )

    def _end_param(self):
        self.stream.write(
            b"""\
    </Parameters>
  </BlastOutput_param>
"""
        )

    def _start_iterations(self):
        self.stream.write(b"<BlastOutput_iterations>\n")

    def _end_iterations(self):
        self.stream.write(b"</BlastOutput_iterations>\n")

    def _start_mbstat(self):
        self.stream.write(b"  <BlastOutput_mbstat>\n")

    def _end_mbstat(self):
        self.stream.write(b"  </BlastOutput_mbstat>\n")

    def _start_iteration(self, record):
        self.stream.write(b"<Iteration>\n")
        self._write_iteration_num(record.num)

    def _end_iteration(self):
        self.stream.write(b"</Iteration>\n")

    def _write_iteration_num(self, num):
        self.stream.write(b"  <Iteration_iter-num>%d</Iteration_iter-num>\n" % num)

    def _write_iteration_query_id(self, query_id):
        self.stream.write(b"<Iteration_query-ID>%b</Iteration_query-ID>\n" % query_id)

    def _write_iteration_query_def(self, query_def):
        self.stream.write(
            b"<Iteration_query-def>%b</Iteration_query-def>\n" % query_def
        )

    def _write_iteration_query_len(self, query_len):
        self.stream.write(
            b"<Iteration_query-len>%d</Iteration_query-len>\n" % query_len
        )

    def _start_iteration_hits(self):
        self.stream.write(b"<Iteration_hits>\n")

    def _end_iteration_hits(self):
        self.stream.write(b"</Iteration_hits>\n")

    def _start_iteration_stat(self):
        self.stream.write(b"  <Iteration_stat>\n")

    def _end_iteration_stat(self):
        self.stream.write(b"  </Iteration_stat>\n")

    def _write_parameters_matrix(self, value):
        self.stream.write(b"      <Parameters_matrix>%b</Parameters_matrix>\n" % value)

    def _write_parameters_expect(self, value):
        self.stream.write(b"      <Parameters_expect>%g</Parameters_expect>\n" % value)

    def _write_parameters_include(self, value):
        self.stream.write(
            b"      <Parameters_include>%g</Parameters_include>\n" % value
        )

    def _write_parameters_sc_match(self, value):
        self.stream.write(
            b"      <Parameters_sc-match>%d</Parameters_sc-match>\n" % value
        )

    def _write_parameters_sc_mismatch(self, value):
        self.stream.write(
            b"      <Parameters_sc-mismatch>%d</Parameters_sc-mismatch>\n" % value
        )

    def _write_parameters_gap_open(self, value):
        self.stream.write(
            b"      <Parameters_gap-open>%d</Parameters_gap-open>\n" % value
        )

    def _write_parameters_gap_extend(self, value):
        self.stream.write(
            b"      <Parameters_gap-extend>%d</Parameters_gap-extend>\n" % value
        )

    def _write_parameters_filter(self, value):
        self.stream.write(b"      <Parameters_filter>%b</Parameters_filter>\n" % value)

    def _write_parameters_pattern(self, value):
        self.stream.write(
            b"      <Parameters_pattern>%b</Parameters_pattern>\n" % value
        )

    def _write_parameters_entrez_query(self, value):
        self.stream.write(
            b"      <Parameters_entrez-query>%b</Parameters_entrez-query>\n" % value
        )

    def _write_statistics_db_num(self, db_num):
        self.stream.write(b"      <Statistics_db-num>%d</Statistics_db-num>\n" % db_num)

    def _write_statistics_db_len(self, db_len):
        self.stream.write(b"      <Statistics_db-len>%d</Statistics_db-len>\n" % db_len)

    def _write_statistics_hsp_len(self, hsp_len):
        self.stream.write(
            b"      <Statistics_hsp-len>%d</Statistics_hsp-len>\n" % hsp_len
        )

    def _write_statistics_eff_space(self, value):
        self.stream.write(
            b"      <Statistics_eff-space>%s</Statistics_eff-space>\n" % value
        )

    def _write_statistics_kappa(self, value):
        self.stream.write(b"      <Statistics_kappa>%r</Statistics_kappa>\n" % value)

    def _write_statistics_lambda(self, value):
        self.stream.write(b"      <Statistics_lambda>%r</Statistics_lambda>\n" % value)

    def _write_statistics_entropy(self, value):
        self.stream.write(
            b"      <Statistics_entropy>%r</Statistics_entropy>\n" % value
        )

    def _start_iteration_hit(self):
        self.stream.write(b"<Hit>\n")

    def _end_iteration_hit(self):
        self.stream.write(b"</Hit>\n")

    def _write_hit_num(self, num):
        self.stream.write(b"  <Hit_num>%d</Hit_num>\n" % num)

    def _write_hit_id(self, hit_id):
        self.stream.write(b"  <Hit_id>%b</Hit_id>\n" % hit_id)

    def _write_hit_def(self, hit_def):
        self.stream.write(b"  <Hit_def>%b</Hit_def>\n" % hit_def)

    def _write_hit_accession(self, hit_accession):
        self.stream.write(b"  <Hit_accession>%b</Hit_accession>\n" % hit_accession)

    def _write_hit_len(self, hit_length):
        self.stream.write(b"  <Hit_len>%d</Hit_len>\n" % hit_length)

    def _start_hit_hsps(self):
        self.stream.write(b"  <Hit_hsps>\n")

    def _end_hit_hsps(self):
        self.stream.write(b"  </Hit_hsps>\n")

    def _start_hsp(self):
        self.stream.write(b"    <Hsp>\n")

    def _end_hsp(self):
        self.stream.write(b"    </Hsp>\n")

    def _write_hsp_num(self, num):
        self.stream.write(b"    <Hsp_num>%d</Hsp_num>\n" % num)

    def _write_hsp_bit_score(self, bit_score):
        self.stream.write(b"    <Hsp_bit-score>%b</Hsp_bit-score>\n" % bit_score)

    def _write_hsp_score(self, score):
        self.stream.write(b"    <Hsp_score>%d</Hsp_score>\n" % score)

    def _write_hsp_evalue(self, evalue):
        self.stream.write(b"    <Hsp_evalue>%b</Hsp_evalue>\n" % evalue)

    def _write_hsp_query_from(self, query_from):
        self.stream.write(b"     <Hsp_query-from>%d</Hsp_query-from>\n" % query_from)

    def _write_hsp_query_to(self, query_to):
        self.stream.write(b"     <Hsp_query-to>%d</Hsp_query-to>\n" % query_to)

    def _write_hsp_hit_from(self, hit_from):
        self.stream.write(b"     <Hsp_hit-from>%d</Hsp_hit-from>\n" % hit_from)

    def _write_hsp_hit_to(self, hit_to):
        self.stream.write(b"     <Hsp_hit-to>%d</Hsp_hit-to>\n" % hit_to)

    def _write_hsp_pattern_from(self, pattern_from):
        self.stream.write(
            b"     <Hsp_pattern-from>%d</Hsp_pattern-from>\n" % pattern_from
        )

    def _write_hsp_pattern_to(self, pattern_to):
        self.stream.write(b"     <Hsp_pattern-to>%d</Hsp_pattern-to>\n" % pattern_to)

    def _write_hsp_query_frame(self, query_frame):
        self.stream.write(b"     <Hsp_query-frame>%d</Hsp_query-frame>\n" % query_frame)

    def _write_hsp_hit_frame(self, hit_frame):
        self.stream.write(b"     <Hsp_hit-frame>%d</Hsp_hit-frame>\n" % hit_frame)

    def _write_hsp_identity(self, identity):
        self.stream.write(b"     <Hsp_identity>%d</Hsp_identity>\n" % identity)

    def _write_hsp_positive(self, positive):
        self.stream.write(b"     <Hsp_positive>%d</Hsp_positive>\n" % positive)

    def _write_hsp_gaps(self, gaps):
        self.stream.write(b"     <Hsp_gaps>%d</Hsp_gaps>\n" % gaps)

    def _write_hsp_align_len(self, align_len):
        self.stream.write(b"     <Hsp_align-len>%d</Hsp_align-len>\n" % align_len)

    def _write_hsp_density(self, density):
        self.stream.write(b"     <Hsp_density>%d</Hsp_density>\n" % density)

    def _write_hsp_qseq(self, qseq):
        self.stream.write(b"      <Hsp_qseq>%b</Hsp_qseq>\n" % qseq)

    def _write_hsp_hseq(self, hseq):
        self.stream.write(b"      <Hsp_hseq>%b</Hsp_hseq>\n" % hseq)

    def _write_hsp_midline(self, midline):
        self.stream.write(b"      <Hsp_midline>%b</Hsp_midline>\n" % midline)


class XML2Writer(BaseXMLWriter):
    """XML2 Writer."""

    def _start_blastoutput(self):
        self.stream.write(
            b"""\
<BlastXML2
    xmlns="http://www.ncbi.nlm.nih.gov"
    xmlns:xs="http://www.w3.org/2001/XMLSchema-instance"
    xs:schemaLocation="http://www.ncbi.nlm.nih.gov http://www.ncbi.nlm.nih.gov/data_specs/schema_alt/NCBI_BlastOutput2.xsd"
>
<BlastOutput2>
  <report>
    <Report>
"""
        )

    def _end_blastoutput(self):
        self.stream.write(
            b"""\
    </Report>
  </report>
</BlastOutput2>
</BlastXML2>
"""
        )

    def _write_program(self, program):
        self.stream.write(b"      <program>%b</program>\n" % program)

    def _write_version(self, version):
        self.stream.write(b"      <version>%b</version>\n" % version)

    def _write_reference(self, reference):
        self.stream.write(b"      <reference>%b</reference>\n" % reference)

    def _write_db(self, db):
        self.stream.write(
            b"""\
      <search-target>
        <Target>
          <db>%b</db>
        </Target>
      </search-target>
"""
            % db
        )

    def _start_param(self):
        self.stream.write(
            b"""\
      <params>
        <Parameters>
"""
        )

    def _end_param(self):
        self.stream.write(
            b"""\
        </Parameters>
      </params>
"""
        )

    def _start_iterations(self):
        self.stream.write(
            b"""\
      <results>
        <Results>
"""
        )

    def _end_iterations(self):
        self.stream.write(
            b"""\
        </Results>
      </results>
"""
        )

    def _start_iteration(self, record):
        try:
            num = record.num
        except AttributeError:
            pass
        else:  # PSIBLAST
            self.stream.write(
                b"""\
            <Iteration>
"""
            )
            self._write_iteration_num(num)
        self.stream.write(
            b"""\
      <search>
        <Search>
"""
        )

    def _end_iteration(self):
        self.stream.write(
            b"""\
        </Search>
      </search>
"""
        )
        if self._program == "psiblast":
            self.stream.write(
                b"""\
            </Iteration>
"""
            )

    def _write_iteration_num(self, num):
        self.stream.write(b"              <iter-num>%d</iter-num>\n" % num)

    def _write_iteration_query_id(self, query_id):
        self.stream.write(
            b"""\
              <query-id>%b</query-id>
"""
            % query_id
        )

    def _write_iteration_query_def(self, query_def):
        self.stream.write(
            b"""\
              <query-title>%b</query-title>
"""
            % query_def
        )

    def _write_iteration_query_len(self, query_len):
        self.stream.write(
            b"""\
              <query-len>%d</query-len>
"""
            % query_len
        )

    def _write_iteration_query_masking(self, location):
        self.stream.write(
            b"""\
              <query-masking>
                <Range>
                  <from>%d</from>
                  <to>%d</to>
                </Range>
              </query-masking>
"""
            % (location.start + 1, location.end)
        )

    def _start_iteration_hits(self):
        self.stream.write(
            b"""\
              <hits>
"""
        )

    def _end_iteration_hits(self):
        self.stream.write(
            b"""\
              </hits>
"""
        )

    def _start_iteration_stat(self):
        self.stream.write(
            b"""\
              <stat>
"""
        )

    def _end_iteration_stat(self):
        self.stream.write(
            b"""\
              </stat>
"""
        )

    def _write_parameters_matrix(self, value):
        self.stream.write(
            b"""\
          <matrix>%b</matrix>
"""
            % value
        )

    def _write_parameters_expect(self, value):
        self.stream.write(
            b"""\
          <expect>%g</expect>
"""
            % value
        )

    def _write_parameters_include(self, value):
        self.stream.write(
            b"""\
          <include>%g</include>
"""
            % value
        )

    def _write_parameters_sc_match(self, value):
        self.stream.write(
            b"""\
          <sc-match>%d</sc-match>
"""
            % value
        )

    def _write_parameters_sc_mismatch(self, value):
        self.stream.write(
            b"""\
          <sc-mismatch>%d</sc-mismatch>
"""
            % value
        )

    def _write_parameters_gap_open(self, value):
        self.stream.write(
            b"""\
          <gap-open>%d</gap-open>
"""
            % value
        )

    def _write_parameters_gap_extend(self, value):
        self.stream.write(
            b"""\
          <gap-extend>%d</gap-extend>
"""
            % value
        )

    def _write_parameters_filter(self, value):
        self.stream.write(
            b"""\
          <filter>%b</filter>
"""
            % value
        )

    def _write_parameters_pattern(self, value):
        self.stream.write(
            b"""\
          <pattern>%b</pattern>
"""
            % value
        )

    def _write_parameters_entrez_query(self, value):
        self.stream.write(
            b"""\
          <entrez-query>%b</entrez-query>
"""
            % value
        )

    def _write_parameters_cbs(self, value):
        self.stream.write(
            b"""\
          <cbs>%d</cbs>
"""
            % value
        )

    def _write_parameters_query_gencode(self, value):
        self.stream.write(
            b"""\
          <query-gencode>%d</query-gencode>
"""
            % value
        )

    def _write_parameters_db_gencode(self, value):
        self.stream.write(
            b"""\
          <db-gencode>%d</db-gencode>
"""
            % value
        )

    def _write_parameters_bl2seq_mode(self, value):
        self.stream.write(
            b"""\
          <bl2seq-mode>%b</bl2seq-mode>
"""
            % value
        )

    def _write_statistics_db_num(self, db_num):
        self.stream.write(
            b"""\
                  <db-num>%d</db-num>
"""
            % db_num
        )

    def _write_statistics_db_len(self, db_len):
        self.stream.write(
            b"""\
                  <db-len>%d</db-len>
"""
            % db_len
        )

    def _write_statistics_hsp_len(self, hsp_len):
        self.stream.write(
            b"""\
                  <hsp-len>%d</hsp-len>
"""
            % hsp_len
        )

    def _write_statistics_eff_space(self, value):
        self.stream.write(
            b"""\
                  <eff-space>%s</eff-space>
"""
            % value
        )

    def _write_statistics_kappa(self, value):
        self.stream.write(
            b"""\
                  <kappa>%r</kappa>
"""
            % value
        )

    def _write_statistics_lambda(self, value):
        self.stream.write(
            b"""\
                  <lambda>%r</lambda>
"""
            % value
        )

    def _write_statistics_entropy(self, value):
        self.stream.write(
            b"""\
                  <entropy>%r</entropy>
"""
            % value
        )

    def _start_iteration_hit(self):
        self.stream.write(
            b"""\
                <Hit>
"""
        )

    def _end_iteration_hit(self):
        self.stream.write(
            b"""\
                </Hit>
"""
        )

    def _write_hit_num(self, num):
        self.stream.write(
            b"""\
                  <num>%d</num>
"""
            % num
        )

    def _start_hit_targets(self):
        self.stream.write(
            b"""\
                  <description>
"""
        )

    def _end_hit_targets(self):
        self.stream.write(
            b"""\
                  </description>
"""
        )

    def _start_hitdescr(self):
        self.stream.write(
            b"""\
                    <HitDescr>
"""
        )

    def _end_hitdescr(self):
        self.stream.write(
            b"""\
                    </HitDescr>
"""
        )

    def _write_hit_id(self, hit_id):
        self.stream.write(
            b"""\
                      <id>%b</id>
"""
            % hit_id
        )

    def _write_hit_def(self, hit_def):
        self.stream.write(
            b"""\
                      <title>%b</title>
"""
            % hit_def
        )

    def _write_hit_accession(self, hit_accession):
        self.stream.write(
            b"""\
                      <accession>%b</accession>
"""
            % hit_accession
        )

    def _write_hit_len(self, hit_length):
        self.stream.write(
            b"""\
                  <len>%d</len>
"""
            % hit_length
        )

    def _write_hit_taxid(self, taxid):
        self.stream.write(
            b"""\
                      <taxid>%d</taxid>
"""
            % taxid
        )

    def _write_hit_sciname(self, sciname):
        self.stream.write(
            b"""\
                      <sciname>%b</sciname>
"""
            % sciname
        )

    def _start_hit_hsps(self):
        self.stream.write(
            b"""\
                  <hsps>
"""
        )

    def _end_hit_hsps(self):
        self.stream.write(
            b"""\
                  </hsps>
"""
        )

    def _start_hsp(self):
        self.stream.write(
            b"""\
                    <Hsp>
"""
        )

    def _end_hsp(self):
        self.stream.write(
            b"""\
                    </Hsp>
"""
        )

    def _write_hsp_num(self, num):
        self.stream.write(
            b"""\
                      <num>%d</num>
"""
            % num
        )

    def _write_hsp_bit_score(self, bit_score):
        self.stream.write(
            b"""\
                      <bit-score>%b</bit-score>
"""
            % bit_score
        )

    def _write_hsp_score(self, score):
        self.stream.write(
            b"""\
                      <score>%d</score>
"""
            % score
        )

    def _write_hsp_evalue(self, evalue):
        self.stream.write(
            b"""\
                      <evalue>%b</evalue>
"""
            % evalue
        )

    def _write_hsp_query_from(self, query_from):
        self.stream.write(
            b"""\
                      <query-from>%d</query-from>
"""
            % query_from
        )

    def _write_hsp_query_to(self, query_to):
        self.stream.write(
            b"""\
                      <query-to>%d</query-to>
"""
            % query_to
        )

    def _write_hsp_hit_from(self, hit_from):
        self.stream.write(
            b"""\
                      <hit-from>%d</hit-from>
"""
            % hit_from
        )

    def _write_hsp_hit_to(self, hit_to):
        self.stream.write(
            b"""\
                      <hit-to>%d</hit-to>
"""
            % hit_to
        )

    def _write_hsp_pattern_from(self, pattern_from):
        self.stream.write(
            b"""\
                      <pattern-from>%d</pattern-from>
"""
            % pattern_from
        )

    def _write_hsp_pattern_to(self, pattern_to):
        self.stream.write(
            b"""\
                      <pattern-to>%d</pattern-to>
"""
            % pattern_to
        )

    def _write_hsp_query_frame(self, query_frame):
        self.stream.write(
            b"""\
                      <query-frame>%d</query-frame>
"""
            % query_frame
        )

    def _write_hsp_hit_frame(self, hit_frame):
        self.stream.write(
            b"""\
                      <hit-frame>%d</hit-frame>
"""
            % hit_frame
        )

    def _write_hsp_identity(self, identity):
        self.stream.write(
            b"""\
                      <identity>%d</identity>
"""
            % identity
        )

    def _write_hsp_positive(self, positive):
        self.stream.write(
            b"""\
                      <positive>%d</positive>
"""
            % positive
        )

    def _write_hsp_gaps(self, gaps):
        self.stream.write(
            b"""\
                      <gaps>%d</gaps>
"""
            % gaps
        )

    def _write_hsp_align_len(self, align_len):
        self.stream.write(
            b"""\
                      <align-len>%d</align-len>
"""
            % align_len
        )

    def _write_hsp_density(self, density):
        self.stream.write(
            b"""\
                      <density>%d</density>
"""
            % density
        )

    def _write_hsp_qseq(self, qseq):
        self.stream.write(
            b"""\
                      <qseq>%b</qseq>
"""
            % qseq
        )

    def _write_hsp_hseq(self, hseq):
        self.stream.write(
            b"""\
                      <hseq>%b</hseq>
"""
            % hseq
        )

    def _write_hsp_midline(self, midline):
        self.stream.write(
            b"""\
                      <midline>%b</midline>
"""
            % midline
        )



# Copyright 2006-2017,2020 by Peter Cock.  All rights reserved.
#
# This file is part of the Biopython distribution and governed by your
# choice of the "Biopython License Agreement" or the "BSD 3-Clause License".
# Please see the LICENSE file that should have been included as part of this
# package.
#
# This module is for reading and writing FASTA format files as SeqRecord
# objects.  The code is partly inspired  by earlier Biopython modules,
# Bio.Fasta.* and the now removed module Bio.SeqIO.FASTA
"""Bio.SeqIO support for the "fasta" (aka FastA or Pearson) file format.

You are expected to use this module via the Bio.SeqIO functions.
"""

from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from Bio import BiopythonDeprecationWarning


from .Interfaces import _clean
from .Interfaces import _get_seq_string
from .Interfaces import _TextIOSource
from .Interfaces import SequenceIterator
from .Interfaces import SequenceWriter

import warnings


def SimpleFastaParser(handle):
    """Iterate over Fasta records as string tuples.

    Arguments:
     - handle - input stream opened in text mode

    For each record a tuple of two strings is returned, the FASTA title
    line (without the leading '>' character), and the sequence (with any
    whitespace removed). The title line is not divided up into an
    identifier (the first word) and comment or description.

    >>> with open("Fasta/dups.fasta") as handle:
    ...     for values in SimpleFastaParser(handle):
    ...         print(values)
    ...
    ('alpha', 'ACGTA')
    ('beta', 'CGTC')
    ('gamma', 'CCGCC')
    ('alpha (again - this is a duplicate entry to test the indexing code)', 'ACGTA')
    ('delta', 'CGCGC')

    """
    # Skip any text before the first record (e.g. blank lines, comments)
    for line in handle:
        if line[0] == ">":
            title = line[1:].rstrip()
            break
    else:
        # no break encountered - probably an empty file
        return

    # Main logic
    # Note, remove trailing whitespace, and any internal spaces
    # (and any embedded \r which are possible in mangled files
    # when not opened in universal read lines mode)
    lines = []
    for line in handle:
        if line[0] == ">":
            yield title, "".join(lines).replace(" ", "").replace("\r", "")
            lines = []
            title = line[1:].rstrip()
            continue
        lines.append(line.rstrip())

    yield title, "".join(lines).replace(" ", "").replace("\r", "")


def FastaTwoLineParser(handle):
    """Iterate over no-wrapping Fasta records as string tuples.

    Arguments:
     - handle - input stream opened in text mode

    Functionally the same as SimpleFastaParser but with a strict
    interpretation of the FASTA format as exactly two lines per
    record, the greater-than-sign identifier with description,
    and the sequence with no line wrapping.

    Any line wrapping will raise an exception, as will excess blank
    lines (other than the special case of a zero-length sequence
    as the second line of a record).

    Examples
    --------
    This file uses two lines per FASTA record:

    >>> with open("Fasta/aster_no_wrap.pro") as handle:
    ...     for title, seq in FastaTwoLineParser(handle):
    ...         print("%s = %s..." % (title, seq[:3]))
    ...
    gi|3298468|dbj|BAA31520.1| SAMIPF = GGH...

    This equivalent file uses line wrapping:

    >>> with open("Fasta/aster.pro") as handle:
    ...     for title, seq in FastaTwoLineParser(handle):
    ...         print("%s = %s..." % (title, seq[:3]))
    ...
    Traceback (most recent call last):
       ...
    ValueError: Expected FASTA record starting with '>' character. Perhaps this file is using FASTA line wrapping? Got: 'MTFGLVYTVYATAIDPKKGSLGTIAPIAIGFIVGANI'

    """
    idx = -1  # for empty file
    for idx, line in enumerate(handle):
        if idx % 2 == 0:  # title line
            if line[0] != ">":
                raise ValueError(
                    "Expected FASTA record starting with '>' character. "
                    "Perhaps this file is using FASTA line wrapping? "
                    f"Got: '{line}'"
                )
            title = line[1:].rstrip()
        else:  # sequence line
            if line[0] == ">":
                raise ValueError(
                    "Two '>' FASTA lines in a row. Missing sequence line "
                    "if this is strict two-line-per-record FASTA format. "
                    f"Have '>{title}' and '{line}'"
                )
            yield title, line.strip()

    if idx == -1:
        pass  # empty file
    elif idx % 2 == 0:  # on a title line
        raise ValueError(
            "Missing sequence line at end of file if this is strict "
            f"two-line-per-record FASTA format. Have title line '{line}'"
        )
    else:
        assert line[0] != ">", "line[0] == '>' ; this should be impossible!"


class FastaIterator(SequenceIterator):
    """Parser for plain Fasta files without comments."""

    modes = "t"

    def __init__(
        self,
        source: _TextIOSource,
        alphabet: None = None,
    ) -> None:
        """Iterate over Fasta records as SeqRecord objects.

        Arguments:
         - source - input stream opened in text mode, or a path to a file
         - alphabet - optional alphabet, not used. Leave as None.

        This parser expects a plain Fasta format without comments or header
        lines.

        By default this will act like calling Bio.SeqIO.parse(handle, "fasta")
        with no custom handling of the title lines:

        >>> with open("Fasta/dups.fasta") as handle:
        ...     for record in FastaIterator(handle):
        ...         print(record.id)
        ...
        alpha
        beta
        gamma
        alpha
        delta

        If you want to modify the records before writing, for example to change
        the ID of each record, you can use a generator function as follows:

        >>> def modify_records(records):
        ...     for record in records:
        ...         record.id = record.id.upper()
        ...         yield record
        ...
        >>> with open('Fasta/dups.fasta') as handle:
        ...     for record in modify_records(FastaIterator(handle)):
        ...         print(record.id)
        ...
        ALPHA
        BETA
        GAMMA
        ALPHA
        DELTA

        """
        if alphabet is not None:
            raise ValueError("The alphabet argument is no longer supported")
        super().__init__(source, fmt="Fasta")
        line = self.stream.readline()
        if not line:
            line = None
        else:
            if not line.startswith(">"):
                raise ValueError(
                    """\
This FASTA file contains comments at the beginning of the file, which are not
allowed by the 'fasta' parser.

To parse this file, you have three options:

(1) Modify your FASTA file to remove such comments at the beginning of the
file.

(2) Use SeqIO.parse with the 'fasta-pearson' format instead of 'fasta'. This
format is consistent with the FASTA format defined by William Pearson's FASTA
aligner software. This format allows for comments before the first sequence;
lines starting with the ';' character anywhere in the file are also regarded
as comment lines and are ignored.

(3) Use the 'fasta-blast' format. This format regards any lines "starting with
'!', '#', or ';' as comment lines. The 'fasta-blast' format may be safer than
the 'fasta-pearson' format, as it explicitly indicates which lines are comments.
"""
                )
        self._line = line

    def __next__(self):
        line = self._line
        if line is None:
            raise StopIteration
        title = line[1:].rstrip()
        # Main logic
        # Note, remove trailing whitespace, and any internal spaces
        # (and any embedded \r which are possible in mangled files
        # when not opened in universal read lines mode)
        lines = []
        for line in self.stream:
            if line[0] == ">":
                break
            lines.append(line)
        else:
            line = None
        self._line = line
        sequence = "".join(lines).encode().translate(None, b" \t\r\n")
        try:
            first_word = title.split(None, 1)[0]
        except IndexError:
            assert not title, repr(title)
            # Should we use SeqRecord default for no ID?
            first_word = ""
        return SeqRecord._from_validated(
            Seq(sequence), id=first_word, name=first_word, description=title
        )


class FastaTwoLineIterator(SequenceIterator):
    """Parser for Fasta files with exactly two lines per record."""

    modes = "t"

    def __init__(self, source):
        """Iterate over two-line Fasta records (as SeqRecord objects).

        Arguments:
         - source - input stream opened in text mode, or a path to a file

        This uses a strict interpretation of the FASTA as requiring
        exactly two lines per record (no line wrapping).

        Only the default title to ID/name/description parsing offered
        by the relaxed FASTA parser is offered.
        """
        super().__init__(source, fmt="FASTA")
        self._data = FastaTwoLineParser(self.stream)

    def __next__(self):
        try:
            title, sequence = next(self._data)
        except StopIteration:
            raise StopIteration from None
        try:
            first_word = title.split(None, 1)[0]
        except IndexError:
            assert not title, repr(title)
            # Should we use SeqRecord default for no ID?
            first_word = ""
        return SeqRecord(
            Seq(sequence), id=first_word, name=first_word, description=title
        )


class FastaBlastIterator(SequenceIterator):
    """Parser for Fasta files, allowing for comments as in BLAST."""

    modes = "t"

    def __init__(
        self,
        source: _TextIOSource,
        alphabet: None = None,
    ) -> None:
        """Iterate over Fasta records as SeqRecord objects.

        Arguments:
         - source - input stream opened in text mode, or a path to a file
         - alphabet - optional alphabet, not used. Leave as None.

        This parser expects the data to be in FASTA format. As in BLAST, lines
        starting with '#', '!', or ';' are interpreted as comments and ignored.

        This iterator acts like calling Bio.SeqIO.parse(handle, "fasta-blast")
        with no custom handling of the title lines:

        >>> with open("Fasta/dups.fasta") as handle:
        ...     for record in FastaIterator(handle):
        ...         print(record.id)
        ...
        alpha
        beta
        gamma
        alpha
        delta

        If you want to modify the records before writing, for example to change
        the ID of each record, you can use a generator function as follows:

        >>> def modify_records(records):
        ...     for record in records:
        ...         record.id = record.id.upper()
        ...         yield record
        ...
        >>> with open('Fasta/dups.fasta') as handle:
        ...     for record in modify_records(FastaIterator(handle)):
        ...         print(record.id)
        ...
        ALPHA
        BETA
        GAMMA
        ALPHA
        DELTA

        """
        if alphabet is not None:
            raise ValueError("The alphabet argument is no longer supported")
        super().__init__(source, fmt="FASTA")
        for line in self.stream:
            if line[0] not in "#!;":
                if not line.startswith(">"):
                    raise ValueError(
                        "Expected FASTA record starting with '>' character.\n"
                        "If this line is a comment, please use '#', '!', or ';' as "
                        "the first character, or use the 'fasta-pearson' "
                        "format for parsing.\n"
                        f"Got: '{line}'"
                    )
                self._line = line
                break
        else:
            self._line = None

    def __next__(self):
        line = self._line
        if line is None:
            raise StopIteration
        title = line[1:].rstrip()
        lines = []
        for line in self.stream:
            # Main logic
            # Note, remove trailing whitespace, and any internal spaces
            # (and any embedded \r which are possible in mangled files
            # when not opened in universal read lines mode)
            if line[0] in "#!;":
                pass
            elif line[0] == ">":
                self_line = line
                break
            else:
                lines.append(line.rstrip())
        else:
            self._line = None
        try:
            first_word = title.split(None, 1)[0]
        except IndexError:
            first_word = ""
        sequence = "".join(lines).replace(" ", "").replace("\r", "")
        return SeqRecord(
            Seq(sequence), id=first_word, name=first_word, description=title
        )


class FastaPearsonIterator(SequenceIterator):
    """Parser for Fasta files, allowing for comments as in the FASTA aligner."""

    modes = "t"

    def __init__(
        self,
        source: _TextIOSource,
        alphabet: None = None,
    ) -> None:
        """Iterate over Fasta records as SeqRecord objects.

        Arguments:
         - source - input stream opened in text mode, or a path to a file
         - alphabet - optional alphabet, not used. Leave as None.

        This parser expects a Fasta format allowing for a header (before the
        first sequence record) and comments (lines starting with ';') as in
        William Pearson's FASTA aligner software.

        This iterator acts as calling Bio.SeqIO.parse(handle, "fasta-pearson")
        with no custom handling of the title lines:

        >>> with open("Fasta/dups.fasta") as handle:
        ...     for record in FastaIterator(handle):
        ...         print(record.id)
        ...
        alpha
        beta
        gamma
        alpha
        delta

        If you want to modify the records before writing, for example to change
        the ID of each record, you can use a generator function as follows:

        >>> def modify_records(records):
        ...     for record in records:
        ...         record.id = record.id.upper()
        ...         yield record
        ...
        >>> with open('Fasta/dups.fasta') as handle:
        ...     for record in modify_records(FastaIterator(handle)):
        ...         print(record.id)
        ...
        ALPHA
        BETA
        GAMMA
        ALPHA
        DELTA

        """
        if alphabet is not None:
            raise ValueError("The alphabet argument is no longer supported")
        super().__init__(source, fmt="Fasta")
        for line in self.stream:
            if line.startswith(">"):
                self._line = line
                break
        else:
            self._line = None

    def __next__(self):
        line = self._line
        if line is None:
            raise StopIteration
        title = line[1:].rstrip()
        lines = []
        for line in self.stream:
            # Main logic
            # Note, remove trailing whitespace, and any internal spaces
            # (and any embedded \r which are possible in mangled files
            # when not opened in universal read lines mode)
            if line[0] == ";":
                pass
            elif line[0] == ">":
                self._line = line
                break
            else:
                lines.append(line.rstrip())
        else:
            self._line = None
        try:
            first_word = title.split(None, 1)[0]
        except IndexError:
            first_word = ""
        sequence = "".join(lines).replace(" ", "").replace("\r", "")
        return SeqRecord(
            Seq(sequence), id=first_word, name=first_word, description=title
        )


class FastaWriter(SequenceWriter):
    """FASTA file writer."""

    modes = "t"

    def __init__(self, target, wrap=60, record2title=None):
        """Create a Fasta writer.

        Arguments:
         - target - Output stream opened in text mode, or a path to a file.
         - wrap -   Optional line length used to wrap sequence lines.
           Defaults to wrapping the sequence at 60 characters
           Use zero (or None) for no wrapping, giving a single
           long line for the sequence.
         - record2title - Optional function to return the text to be
           used for the title line of each record.  By default
           a combination of the record.id and record.description
           is used.  If the record.description starts with the
           record.id, then just the record.description is used.

        You can either use::

            handle = open(filename, "w")
            writer = FastaWriter(handle)
            writer.write_file(myRecords)
            handle.close()

        Or, follow the sequential file writer system, for example::

            handle = open(filename, "w")
            writer = FastaWriter(handle)
            ...
            Multiple writer.write_record() and/or writer.write_records() calls
            ...
            handle.close()

        """
        super().__init__(target)
        if wrap:
            if wrap < 1:
                raise ValueError
        self.wrap = wrap
        self.record2title = record2title

    @classmethod
    def to_string(cls, record):
        """Turn a SeqRecord into a FASTA formatted string, and return it."""
        id = _clean(record.id)
        description = _clean(record.description)
        if description and description.split(None, 1)[0] == id:
            # The description includes the id at the start
            title = description
        elif description:
            title = f"{id} {description}"
        else:
            title = id
        assert "\n" not in title
        assert "\r" not in title
        lines = [f">{title}\n"]

        data = _get_seq_string(record)  # Catches sequence being None
        assert "\n" not in data
        assert "\r" not in data
        for i in range(0, len(data), 60):
            lines.append(data[i : i + 60] + "\n")

        return "".join(lines)

    def write_record(self, record):
        """Write a single Fasta record to the file."""
        if self.record2title:
            title = self.clean(self.record2title(record))
        else:
            id = self.clean(record.id)
            description = self.clean(record.description)
            if description and description.split(None, 1)[0] == id:
                # The description includes the id at the start
                title = description
            elif description:
                title = f"{id} {description}"
            else:
                title = id

        assert "\n" not in title
        assert "\r" not in title
        self.handle.write(f">{title}\n")

        data = _get_seq_string(record)  # Catches sequence being None

        assert "\n" not in data
        assert "\r" not in data

        if self.wrap:
            for i in range(0, len(data), self.wrap):
                self.handle.write(data[i : i + self.wrap] + "\n")
        else:
            self.handle.write(data + "\n")


class FastaTwoLineWriter(FastaWriter):
    """Class to write 2-line per record Fasta format files.

    This means we write the sequence information  without line
    wrapping, and will always write a blank line for an empty
    sequence.
    """

    def __init__(self, handle, record2title=None):
        """Create a 2-line per record Fasta writer.

        Arguments:
         - handle - Handle to an output file, e.g. as returned
           by open(filename, "w")
         - record2title - Optional function to return the text to be
           used for the title line of each record.  By default
           a combination of the record.id and record.description
           is used.  If the record.description starts with the
           record.id, then just the record.description is used.

        You can either use::

            handle = open(filename, "w")
            writer = FastaWriter(handle)
            writer.write_file(myRecords)
            handle.close()

        Or, follow the sequential file writer system, for example::

            handle = open(filename, "w")
            writer = FastaWriter(handle)
            ...
            Multiple writer.write_record() and/or writer.write_records() calls
            ...
            handle.close()

        """
        super().__init__(handle, wrap=None, record2title=record2title)

    @classmethod
    def to_string(cls, record):
        """Return a string in FASTA format with the sequence as one line."""
        id = _clean(record.id)
        description = _clean(record.description)
        if description and description.split(None, 1)[0] == id:
            # The description includes the id at the start
            title = description
        elif description:
            title = f"{id} {description}"
        else:
            title = id
        assert "\n" not in title
        assert "\r" not in title

        data = _get_seq_string(record)  # Catches sequence being None
        assert "\n" not in data
        assert "\r" not in data

        return f">{title}\n{data}\n"


def as_fasta(record):
    """Turn a SeqRecord into a FASTA formatted string."""
    warnings.warn(
        """\
FastaIO.as_fasta is deprecated.

Instead of

FastaIO.as_fasta(record)

please use

format(record, "fasta")
""",
        DeprecationWarning,
    )
    return FastaWriter.to_string(record)


def as_fasta_2line(record):
    """Turn a SeqRecord into a two-line FASTA formatted string."""
    warnings.warn(
        """\
FastaIO.as_fasta_2line is deprecated.

Instead of

FastaIO.as_fasta_2line(record)

please use

format(record, "fasta-2line")
""",
        DeprecationWarning,
    )
    return FastaTwoLineWriter.to_string(record)


if __name__ == "__main__":
    from Bio._utils import run_doctest

    run_doctest(verbose=0)



# Copyright 2012 by Eric Talevich.  All rights reserved.
#
# This file is part of the Biopython distribution and governed by your
# choice of the "Biopython License Agreement" or the "BSD 3-Clause License".
# Please see the LICENSE file that should have been included as part of this
# package.
"""Bio.SeqIO support for accessing sequences in PDB and mmCIF files."""

import collections
import warnings

from Bio import BiopythonParserWarning
from Bio.Data.PDBData import protein_letters_3to1
from Bio.Data.PDBData import protein_letters_3to1_extended
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord

from .Interfaces import _TextIOSource
from .Interfaces import SequenceIterator

_aa3to1_dict = {}
_aa3to1_dict.update(protein_letters_3to1)
_aa3to1_dict.update(protein_letters_3to1_extended)


def _res2aacode(residue, undef_code="X"):
    """Return the one-letter amino acid code from the residue name.

    Non-amino acid are returned as "X".
    """
    if isinstance(residue, str):
        return _aa3to1_dict.get(residue, undef_code)

    return _aa3to1_dict.get(residue.resname, undef_code)


def AtomIterator(pdb_id, structure):
    """Return SeqRecords from Structure objects.

    Base function for sequence parsers that read structures Bio.PDB parsers.

    Once a parser from Bio.PDB has been used to load a structure into a
    Bio.PDB.Structure.Structure object, there is no difference in how the
    sequence parser interprets the residue sequence. The functions in this
    module may be used by SeqIO modules wishing to parse sequences from lists
    of residues.

    Calling functions must pass a Bio.PDB.Structure.Structure object.


    See Bio.SeqIO.PdbIO.PdbAtomIterator and Bio.SeqIO.PdbIO.CifAtomIterator for
    details.
    """
    model = structure[0]
    for chn_id, chain in sorted(model.child_dict.items()):
        # HETATM mod. res. policy: remove mod if in sequence, else discard
        residues = [
            res
            for res in chain.get_unpacked_list()
            if _res2aacode(res.get_resname().upper()) != "X"
        ]
        if not residues:
            continue
        # Identify missing residues in the structure
        # (fill the sequence with 'X' residues in these regions)
        gaps = []
        rnumbers = [r.id[1] for r in residues]
        for i, rnum in enumerate(rnumbers[:-1]):
            if rnumbers[i + 1] != rnum + 1 and rnumbers[i + 1] != rnum:
                # It's a gap!
                gaps.append((i + 1, rnum, rnumbers[i + 1]))
        if gaps:
            res_out = []
            prev_idx = 0
            for i, pregap, postgap in gaps:
                if postgap > pregap:
                    gapsize = postgap - pregap - 1
                    res_out.extend(_res2aacode(x) for x in residues[prev_idx:i])
                    prev_idx = i
                    res_out.append("X" * gapsize)
                else:
                    warnings.warn(
                        "Ignoring out-of-order residues after a gap",
                        BiopythonParserWarning,
                    )
                    # Keep the normal part, drop the out-of-order segment
                    # (presumably modified or hetatm residues, e.g. 3BEG)
                    res_out.extend(_res2aacode(x) for x in residues[prev_idx:i])
                    break
            else:
                # Last segment
                res_out.extend(_res2aacode(x) for x in residues[prev_idx:])
        else:
            # No gaps
            res_out = [_res2aacode(x) for x in residues]
        record_id = f"{pdb_id}:{chn_id}"
        # ENH - model number in SeqRecord id if multiple models?
        # id = "Chain%s" % str(chain.id)
        # if len(structure) > 1 :
        #     id = ("Model%s|" % str(model.id)) + id

        record = SeqRecord(Seq("".join(res_out)), id=record_id, description=record_id)
        # TODO: Test PDB files with DNA and RNA too:
        record.annotations["molecule_type"] = "protein"

        record.annotations["model"] = model.id
        record.annotations["chain"] = chain.id

        record.annotations["start"] = int(rnumbers[0])
        record.annotations["end"] = int(rnumbers[-1])
        yield record


class PdbSeqresIterator(SequenceIterator):
    """Parser for PDB files."""

    modes = "t"

    def __init__(self, source: _TextIOSource) -> None:
        """Iterate over chains in a PDB file as SeqRecord objects.

        Arguments:
         - source - input stream opened in text mode, or a path to a file

        The sequences are derived from the SEQRES lines in the
        PDB file header, not the atoms of the 3D structure.

        Specifically, these PDB records are handled: DBREF, DBREF1, DBREF2, SEQADV, SEQRES, MODRES

        See: http://www.wwpdb.org/documentation/format23/sect3.html

        This gets called internally via Bio.SeqIO for the SEQRES based interpretation
        of the PDB file format:

        >>> from Bio import SeqIO
        >>> for record in SeqIO.parse("PDB/1A8O.pdb", "pdb-seqres"):
        ...     print("Record id %s, chain %s" % (record.id, record.annotations["chain"]))
        ...     print(record.dbxrefs)
        ...
        Record id 1A8O:A, chain A
        ['UNP:P12497', 'UNP:POL_HV1N5']

        Equivalently,

        >>> with open("PDB/1A8O.pdb") as handle:
        ...     for record in PdbSeqresIterator(handle):
        ...         print("Record id %s, chain %s" % (record.id, record.annotations["chain"]))
        ...         print(record.dbxrefs)
        ...
        Record id 1A8O:A, chain A
        ['UNP:P12497', 'UNP:POL_HV1N5']

        Note the chain is recorded in the annotations dictionary, and any PDB DBREF
        lines are recorded in the database cross-references list.
        """
        super().__init__(source, fmt="PDB")
        self.cache = None

    def __next__(self):
        """Iterate over the records in the PDB file."""
        if self.cache is None:
            chains = collections.defaultdict(list)
            metadata = collections.defaultdict(list)

            rec_name = None
            for line in self.stream:
                rec_name = line[0:6].strip()
                if rec_name == "SEQRES":
                    # NB: We only actually need chain ID and the residues here;
                    # commented bits are placeholders from the wwPDB spec.
                    # Serial number of the SEQRES record for the current chain.
                    # Starts at 1 and increments by one each line.
                    # Reset to 1 for each chain.
                    # ser_num = int(line[8:10])
                    # Chain identifier. This may be any single legal character,
                    # including a blank which is used if there is only one
                    # chain.
                    chn_id = line[11]
                    # Number of residues in the chain (repeated on every record)
                    # num_res = int(line[13:17])
                    residues = [_res2aacode(res) for res in line[19:].split()]
                    chains[chn_id].extend(residues)
                elif rec_name == "DBREF":
                    #  ID code of this entry (PDB ID)
                    pdb_id = line[7:11]
                    # Chain identifier.
                    chn_id = line[12]
                    # Initial sequence number of the PDB sequence segment.
                    # seq_begin = int(line[14:18])
                    # Initial insertion code of the PDB sequence segment.
                    # icode_begin = line[18]
                    # Ending sequence number of the PDB sequence segment.
                    # seq_end = int(line[20:24])
                    # Ending insertion code of the PDB sequence segment.
                    # icode_end = line[24]
                    # Sequence database name.
                    database = line[26:32].strip()
                    # Sequence database accession code.
                    db_acc = line[33:41].strip()
                    # Sequence database identification code.
                    db_id_code = line[42:54].strip()
                    # Initial sequence number of the database seqment.
                    # db_seq_begin = int(line[55:60])
                    # Insertion code of initial residue of the segment, if PDB
                    # is the reference.
                    # db_icode_begin = line[60]
                    # Ending sequence number of the database segment.
                    # db_seq_end = int(line[62:67])
                    # Insertion code of the ending residue of the segment, if
                    # PDB is the reference.
                    # db_icode_end = line[67]
                    metadata[chn_id].append(
                        {
                            "pdb_id": pdb_id,
                            "database": database,
                            "db_acc": db_acc,
                            "db_id_code": db_id_code,
                        }
                    )
                elif rec_name == "DBREF1":
                    # ID code of this entry (PDB ID)
                    pdb_id = line[7:11]
                    # Chain identifier.
                    chn_id = line[12]
                    # Sequence database name.
                    database = line[26:32].strip()
                    # Sequence database identification code.
                    db_id_code = line[47:67].strip()
                elif rec_name == "DBREF2":
                    # Ensure ID code and chain are consistent:
                    if pdb_id != line[7:11] or chn_id != line[12]:
                        raise ValueError("DBREF2 identifiers do not match")
                    # Sequence database accession code.
                    db_acc = line[18:40].strip()
                    metadata[chn_id].append(
                        {
                            "pdb_id": pdb_id,
                            "database": database,
                            "db_acc": db_acc,
                            "db_id_code": db_id_code,
                        }
                    )
                # ENH: 'SEQADV' 'MODRES'

            if rec_name is None:
                raise ValueError("Empty file.")

            self.cache = []
            for chn_id, residues in sorted(chains.items()):
                record = SeqRecord(Seq("".join(residues)))
                record.annotations = {"chain": chn_id}
                # TODO: Test PDB files with DNA and RNA too:
                record.annotations["molecule_type"] = "protein"
                if chn_id in metadata:
                    m = metadata[chn_id][0]
                    record.id = record.name = f"{m['pdb_id']}:{chn_id}"
                    record.description = (
                        f"{m['database']}:{m['db_acc']} {m['db_id_code']}"
                    )
                    for melem in metadata[chn_id]:
                        record.dbxrefs.extend(
                            [
                                f"{melem['database']}:{melem['db_acc']}",
                                f"{melem['database']}:{melem['db_id_code']}",
                            ]
                        )
                else:
                    record.id = chn_id
                self.cache.append(record)
        try:
            record = self.cache.pop(0)
        except IndexError:
            raise StopIteration
        else:
            return record


class PdbAtomIterator(SequenceIterator):
    """Parser for structures in a PDB files."""

    modes = "t"

    def __init__(self, source: _TextIOSource) -> None:
        """Iterate over structures in a PDB file as SeqRecord objects.

        Argument source is a file-like object or a path to a file.

        The sequences are derived from the 3D structure (ATOM records), not the
        SEQRES lines in the PDB file header.

        Unrecognised three letter amino acid codes (e.g. "CSD") from HETATM
        entries are converted to "X" in the sequence.

        In addition to information from the PDB header (which is the same for
        all records), the following chain specific information is placed in the
        annotation:

        record.annotations["residues"] = List of residue ID strings
        record.annotations["chain"] = Chain ID (typically A, B ,...)
        record.annotations["model"] = Model ID (typically zero)

        Where amino acids are missing from the structure, as indicated by
        residue numbering, the sequence is filled in with 'X' characters to
        match the size of the missing region, and  None is included as the
        corresponding entry in the list record.annotations["residues"].

        This function uses the Bio.PDB module to do most of the hard work. The
        annotation information could be improved but this extra parsing should
        be done in parse_pdb_header, not this module.

        This gets called internally via Bio.SeqIO for the atom based
        interpretation of the PDB file format:

        >>> from Bio import SeqIO
        >>> for record in SeqIO.parse("PDB/1A8O.pdb", "pdb-atom"):
        ...     print("Record id %s, chain %s" % (record.id, record.annotations["chain"]))
        ...
        Record id 1A8O:A, chain A

        Equivalently,

        >>> with open("PDB/1A8O.pdb") as handle:
        ...     for record in PdbAtomIterator(handle):
        ...         print("Record id %s, chain %s" % (record.id, record.annotations["chain"]))
        ...
        Record id 1A8O:A, chain A

        """
        # TODO - Add record.annotations to the doctest, esp the residues (not working?)

        # Only import PDB when needed, to avoid/delay NumPy dependency in SeqIO
        from Bio.PDB.PDBParser import PDBParser

        structure = PDBParser().get_structure(None, source)
        pdb_id = structure.header["idcode"]
        if not pdb_id:
            warnings.warn(
                "'HEADER' line not found; can't determine PDB ID.",
                BiopythonParserWarning,
            )
            pdb_id = "????"

        records = []
        for record in AtomIterator(pdb_id, structure):
            # The PDB header was loaded as a dictionary, so let's reuse it all
            record.annotations.update(structure.header)

            # ENH - add letter annotations -- per-residue info, e.g. numbers

            records.append(record)

        self.records = iter(records)

    def __next__(self):
        return next(self.records)


PDBX_POLY_SEQ_SCHEME_FIELDS = (
    "_pdbx_poly_seq_scheme.asym_id",  # Chain ID
    "_pdbx_poly_seq_scheme.mon_id",  # Residue type
)

STRUCT_REF_FIELDS = (
    "_struct_ref.id",  # ID of this reference
    "_struct_ref.db_name",  # Name of the database
    "_struct_ref.db_code",  # Code for this entity
    "_struct_ref.pdbx_db_accession",  # DB accession ID of ref
)

STRUCT_REF_SEQ_FIELDS = (
    "_struct_ref_seq.ref_id",  # Pointer to _struct_ref
    "_struct_ref_seq.pdbx_PDB_id_code",  # PDB ID of this structure
    "_struct_ref_seq.pdbx_strand_id",  # Chain ID of the reference
)


class CifSeqresIterator(SequenceIterator):
    """Parser for chains in an mmCIF files."""

    modes = "t"

    def __init__(self, source: _TextIOSource) -> None:
        """Iterate over chains in an mmCIF file as SeqRecord objects.

        Argument source is a file-like object or a path to a file.

        The sequences are derived from the _entity_poly_seq entries in the
        mmCIF file, not the atoms of the 3D structure.

        Specifically, these mmCIF records are handled: _pdbx_poly_seq_scheme
        and _struct_ref_seq. The _pdbx_poly_seq records contain sequence
        information, and the _struct_ref_seq records contain database
        cross-references.

        See:
        http://mmcif.wwpdb.org/dictionaries/mmcif_pdbx_v40.dic/Categories/pdbx_poly_seq_scheme.html
        and
        http://mmcif.wwpdb.org/dictionaries/mmcif_pdbx_v50.dic/Categories/struct_ref_seq.html

        This gets called internally via Bio.SeqIO for the sequence-based
        interpretation of the mmCIF file format:

        >>> from Bio import SeqIO
        >>> for record in SeqIO.parse("PDB/1A8O.cif", "cif-seqres"):
        ...     print("Record id %s, chain %s" % (record.id, record.annotations["chain"]))
        ...     print(record.dbxrefs)
        ...
        Record id 1A8O:A, chain A
        ['UNP:P12497', 'UNP:POL_HV1N5']

        Equivalently,

        >>> with open("PDB/1A8O.cif") as handle:
        ...     for record in CifSeqresIterator(handle):
        ...         print("Record id %s, chain %s" % (record.id, record.annotations["chain"]))
        ...         print(record.dbxrefs)
        ...
        Record id 1A8O:A, chain A
        ['UNP:P12497', 'UNP:POL_HV1N5']

        Note the chain is recorded in the annotations dictionary, and any mmCIF
        _struct_ref_seq entries are recorded in the database cross-references
        list.
        """

        # Only import PDB when needed, to avoid/delay NumPy dependency in SeqIO
        from Bio.PDB.MMCIF2Dict import MMCIF2Dict

        chains = collections.defaultdict(list)
        metadata = collections.defaultdict(list)
        records = MMCIF2Dict(source)

        # Explicitly convert records to list (See #1533).
        # If an item is not present, use an empty list
        for field in (
            PDBX_POLY_SEQ_SCHEME_FIELDS + STRUCT_REF_SEQ_FIELDS + STRUCT_REF_FIELDS
        ):
            if field not in records:
                records[field] = []
            elif not isinstance(records[field], list):
                records[field] = [records[field]]

        for asym_id, mon_id in zip(
            records["_pdbx_poly_seq_scheme.asym_id"],
            records["_pdbx_poly_seq_scheme.mon_id"],
        ):
            mon_id_1l = _res2aacode(mon_id)
            chains[asym_id].append(mon_id_1l)

        # Build a dict of _struct_ref records, indexed by the id field:
        struct_refs = {}
        for ref_id, db_name, db_code, db_acc in zip(
            records["_struct_ref.id"],
            records["_struct_ref.db_name"],
            records["_struct_ref.db_code"],
            records["_struct_ref.pdbx_db_accession"],
        ):
            struct_refs[ref_id] = {
                "database": db_name,
                "db_id_code": db_code,
                "db_acc": db_acc,
            }

        # Look through _struct_ref_seq records, look up the corresponding
        # _struct_ref and add an entry to the metadata list for this chain.
        for ref_id, pdb_id, chain_id in zip(
            records["_struct_ref_seq.ref_id"],
            records["_struct_ref_seq.pdbx_PDB_id_code"],
            records["_struct_ref_seq.pdbx_strand_id"],
        ):
            struct_ref = struct_refs[ref_id]

            # The names here mirror those in PdbIO
            metadata[chain_id].append({"pdb_id": pdb_id})
            metadata[chain_id][-1].update(struct_ref)

        records = []  # type: ignore
        for chn_id, residues in sorted(chains.items()):
            record = SeqRecord(Seq("".join(residues)))
            record.annotations = {"chain": chn_id}
            # TODO: Test PDB files with DNA and RNA too:
            record.annotations["molecule_type"] = "protein"
            if chn_id in metadata:
                m = metadata[chn_id][0]
                record.id = record.name = f"{m['pdb_id']}:{chn_id}"
                record.description = f"{m['database']}:{m['db_acc']} {m['db_id_code']}"
                for melem in metadata[chn_id]:
                    record.dbxrefs.extend(
                        [
                            f"{melem['database']}:{melem['db_acc']}",
                            f"{melem['database']}:{melem['db_id_code']}",
                        ]
                    )
            else:
                record.id = chn_id
            records.append(record)  # type: ignore

        self.records = iter(records)

    def __next__(self):
        return next(self.records)


class CifAtomIterator(SequenceIterator):
    """Parser for structures in an mmCIF files."""

    modes = "t"

    def __init__(self, source: _TextIOSource) -> None:
        """Iterate over structures in an mmCIF file as SeqRecord objects.

        Argument source is a file-like object or a path to a file.

        The sequences are derived from the 3D structure (_atom_site.* fields)
        in the mmCIF file.

        Unrecognised three letter amino acid codes (e.g. "CSD") from HETATM
        entries are converted to "X" in the sequence.

        In addition to information from the PDB header (which is the same for
        all records), the following chain specific information is placed in the
        annotation:

        record.annotations["residues"] = List of residue ID strings
        record.annotations["chain"] = Chain ID (typically A, B ,...)
        record.annotations["model"] = Model ID (typically zero)

        Where amino acids are missing from the structure, as indicated by
        residue numbering, the sequence is filled in with 'X' characters to
        match the size of the missing region, and  None is included as the
        corresponding entry in the list record.annotations["residues"].

        This function uses the Bio.PDB module to do most of the hard work. The
        annotation information could be improved but this extra parsing should
        be done in parse_pdb_header, not this module.

        This gets called internally via Bio.SeqIO for the atom based
        interpretation of the PDB file format:

        >>> from Bio import SeqIO
        >>> for record in SeqIO.parse("PDB/1A8O.cif", "cif-atom"):
        ...     print("Record id %s, chain %s" % (record.id, record.annotations["chain"]))
        ...
        Record id 1A8O:A, chain A

        Equivalently,

        >>> with open("PDB/1A8O.cif") as handle:
        ...     for record in CifAtomIterator(handle):
        ...         print("Record id %s, chain %s" % (record.id, record.annotations["chain"]))
        ...
        Record id 1A8O:A, chain A

        """
        # TODO - Add record.annotations to the doctest, esp the residues (not working?)

        # Only import parser when needed, to avoid/delay NumPy dependency in SeqIO
        from Bio.PDB.MMCIFParser import MMCIFParser

        structure = MMCIFParser().get_structure(None, source)
        pdb_id = structure.header["idcode"]
        if not pdb_id:
            warnings.warn("Could not determine the PDB ID.", BiopythonParserWarning)
            pdb_id = "????"
        self.records = AtomIterator(pdb_id, structure)

    def __next__(self):
        return next(self.records)


if __name__ == "__main__":
    from Bio._utils import run_doctest

    run_doctest(verbose=0)



"""A module for interacting with the AlphaFold Protein Structure Database.

See the `database website <https://alphafold.com/>`_ and the `API docs <https://alphafold.com/api-docs/>`_.
"""

import json
import os
from os import PathLike
from collections.abc import Iterator
from typing import Optional
from typing import Union
from urllib.request import urlopen

from Bio.PDB.MMCIFParser import MMCIFParser
from Bio.PDB.Structure import Structure as StructuralModel


def get_predictions(qualifier: str) -> Iterator[dict]:
    """Get all AlphaFold predictions for a UniProt accession.

    :param qualifier: A UniProt accession, e.g. P00520
    :type qualifier: str
    :return: The AlphaFold predictions
    :rtype: Iterator[dict]
    """
    url = f"https://alphafold.com/api/prediction/{qualifier}"
    # Retrieve the AlphaFold predictions with urllib
    with urlopen(url) as response:
        yield from json.loads(response.read().decode())


def _get_mmcif_file_path_for(
    prediction: dict, directory: str | bytes | PathLike | None = None
) -> str:
    """Get the path to the mmCIF file for an AlphaFold prediction.

    :param prediction: An AlphaFold prediction
    :type prediction: dict
    :param directory: The directory that stores the mmCIF file, defaults to the current working directory
    :type directory: Union[int, str, bytes, PathLike], optional
    :return: The path to the mmCIF file
    :rtype: str
    """
    if directory is None:
        directory = os.getcwd()

    cif_url = prediction["cifUrl"]
    # Get the file name from the URL
    file_name = cif_url.split("/")[-1]
    return str(os.path.join(directory, file_name))


def download_cif_for(
    prediction: dict, directory: str | bytes | PathLike | None = None
) -> str:
    """Download the mmCIF file for an AlphaFold prediction.

    Downloads the file to the current working directory if no destination is specified.

    :param prediction: An AlphaFold prediction
    :type prediction: dict
    :param directory: The directory to write the mmCIF data to, defaults to the current working directory
    :type directory: Union[int, str, bytes, PathLike], optional
    :return: The path to the mmCIF file
    :rtype: str
    """
    if directory is None:
        directory = os.getcwd()

    cif_url = prediction["cifUrl"]
    # Create the directory in case it does not exist
    os.makedirs(directory, exist_ok=True)
    file_path = _get_mmcif_file_path_for(prediction, directory)

    if os.path.exists(file_path):
        print(f"File {file_path} already exists, skipping download.")
    else:
        with urlopen(cif_url) as response:
            data = response.read()
        # Write the data to destination
        with open(file_path, "wb") as file:
            file.write(data)

    return file_path


def get_structural_models_for(
    qualifier: str,
    mmcif_parser: MMCIFParser | None = None,
    directory: str | bytes | PathLike | None = None,
) -> Iterator[StructuralModel]:
    """Get the PDB structures for a UniProt accession.

    Downloads the mmCIF files to the directory if they are not present.

    :param qualifier: A UniProt accession, e.g. P00520
    :type qualifier: str
    :param mmcif_parser: The mmCIF parser to use, defaults to ``MMCIFParser()``
    :type mmcif_parser: MMCIFParser, optional
    :param directory: The directory to store the mmCIF data, defaults to the current working directory
    :type directory: Union[int, str, bytes, PathLike], optional
    :return: An iterator over the PDB structures
    :rtype: Iterator[PDBStructure]
    """
    if mmcif_parser is None:
        mmcif_parser = MMCIFParser()
    if directory is None:
        directory = os.getcwd()

    for prediction in get_predictions(qualifier):
        mmcif_path = _get_mmcif_file_path_for(prediction, directory)

        if not os.path.exists(mmcif_path):
            mmcif_path = download_cif_for(prediction, directory)

        yield mmcif_parser.get_structure(qualifier, mmcif_path)



# Copyright 2008-2016 by Peter Cock.  All rights reserved.
#
# This file is part of the Biopython distribution and governed by your
# choice of the "Biopython License Agreement" or the "BSD 3-Clause License".
# Please see the LICENSE file that should have been included as part of this
# package.
"""Bio.Align support for "emboss" alignment output from EMBOSS tools.

This module contains a parser for the EMBOSS srspair/pair/simple file format,
for example from the needle, water, and stretcher tools.
"""

from Bio.Align import Alignment
from Bio.Align import interfaces
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord


class AlignmentIterator(interfaces.AlignmentIterator):
    """Emboss alignment iterator.

    For reading the (pairwise) alignments from EMBOSS tools in what they
    call the "pairs" and "simple" formats.
    """

    fmt = "EMBOSS"

    def _read_header(self, stream):
        line = stream.readline()
        if not line:
            raise ValueError("Empty file.")
        if line.rstrip() != "########################################":
            raise ValueError("Unexpected line: %s") % line

        # assume srspair format (default) if not specified explicitly in
        # the output file
        self.metadata = {}
        self.metadata["Align_format"] = "srspair"
        commandline = None
        for line in stream:
            if line.rstrip() == "########################################":
                break
            if not line.startswith("# "):
                raise ValueError("Unexpected line: %s") % line
            if commandline is not None:
                if line.startswith("#    "):
                    commandline += " " + line[1:].strip()
                    continue
                self.metadata["Command line"] = commandline
                commandline = None
            key, value = line[2:].split(":", 1)
            if key == "Program":
                self.metadata["Program"] = value.strip()
            elif key == "Rundate":
                self.metadata["Rundate"] = value.strip()
            elif key == "Report_file":
                self.metadata["Report_file"] = value.strip()
            elif key == "Align_format":
                self.metadata["Align_format"] = value.strip()
            elif key == "Commandline":
                commandline = value.strip()

    def _read_next_alignment(self, stream):
        number_of_sequences = None
        annotations = {}
        for line in stream:
            line = line.rstrip("\r\n")
            if not line:
                continue
            elif line.startswith("#---------------------------------------"):
                # may appear between alignments
                continue
            elif line.startswith("#======================================="):
                # found the alignment metadata start
                identifiers = []
                ncols = None
                sequences = None
                break
            else:
                raise ValueError("Unexpected line: %s" % line)
        for line in stream:
            line = line.rstrip("\r\n")
            if line == "#=======================================":
                # reached the end of alignment metadata
                break
            elif line.strip() == "#":
                continue
            elif not line.startswith("# "):
                raise ValueError("Unexpected line: %s") % line
            try:
                key, value = line[2:].split(":", 1)
            except ValueError:
                # An equal sign is used for Longest_Identity,
                # Longest_Similarity, Shortest_Identity, and
                # Shortest_Similarity, which are included if command line
                # argument -nobrief was used.
                key, value = line[2:].split(" = ", 1)
            if key == "Aligned_sequences":
                number_of_sequences = int(value.strip())
                assert len(identifiers) == 0
                # Should now expect the record identifiers...
                for i, line in enumerate(stream):
                    if not line.startswith("# "):
                        raise ValueError("Unexpected line: %s") % line
                    number, identifier = line[2:].split(":")
                    assert i + 1 == int(number)
                    identifiers.append(identifier.strip())
                    if len(identifiers) == number_of_sequences:
                        break
            elif key == "Matrix":
                annotations[key] = value.strip()
            elif key == "Gap_penalty":
                annotations[key] = float(value.strip())
            elif key == "Extend_penalty":
                annotations[key] = float(value.strip())
            elif key == "Length":
                ncols = int(value.strip())
            elif key == "Identity":
                annotations[key] = int(value.strip().split("/")[0])
            elif key == "Similarity":
                annotations[key] = int(value.strip().split("/")[0])
            elif key == "Gaps":
                annotations[key] = int(value.strip().split("/")[0])
            elif key == "Score":
                annotations[key] = float(value.strip())
            # TODO:
            # The following are generated if the -nobrief command line
            # argument used. We could simply calculate them from the
            # alignment, but then we have to define what we mean by
            # "similar". For now, simply store them as an annotation.
            elif key == "Longest_Identity":
                annotations[key] = value.strip()
            elif key == "Longest_Similarity":
                annotations[key] = value.strip()
            elif key == "Shortest_Identity":
                annotations[key] = value.strip()
            elif key == "Shortest_Similarity":
                annotations[key] = value.strip()
            else:
                raise ValueError("Failed to parse line '%s'" % line)
        else:
            return
        if len(identifiers) == 0:
            raise ValueError("Number of sequences missing!")
        if ncols is None:
            raise ValueError("Length of alignment missing!")
        sequences = [""] * number_of_sequences
        aligned_sequences = [""] * number_of_sequences
        consensus = ""
        starts = [0] * number_of_sequences
        ends = [0] * number_of_sequences
        column = 0
        index = 0
        for line in stream:
            line = line.rstrip("\r\n")
            # parse the sequences
            if not line:
                # empty line
                if index == number_of_sequences:
                    # reached the end of an alignment block
                    if column == ncols:
                        # reached the end of the sequences
                        break
                    index = 0
                continue
            prefix = line[:21].strip()
            if prefix == "":
                # match line
                consensus += line[21:71]
            else:
                identifier, start = prefix.split(None, 1)
                assert identifiers[index].startswith(identifier)
                aligned_sequence, end = line[21:].split(None, 1)
                start = int(start)
                end = int(end)
                length = len(sequences[index])
                sequence = aligned_sequence.replace("-", "")
                if length == 0 and len(sequence) > 0:
                    if start < end:
                        start -= 1  # Python counting
                        assert end == start + len(sequence)
                    else:
                        end -= 1  # Python counting
                        assert end == start - len(sequence)
                    # Record the start
                    starts[index] = start
                else:
                    if starts[index] <= ends[index]:
                        # forward strand
                        if (
                            self.metadata["Align_format"] == "srspair"
                            and len(sequence) == 0
                        ):
                            assert start == ends[index]
                            assert end == start
                        else:
                            start -= 1
                            assert end == start + len(sequence)
                    else:
                        if (
                            self.metadata["Align_format"] == "srspair"
                            and len(sequence) == 0
                        ):
                            assert start - 1 == ends[index]
                            assert end == start
                        else:
                            end -= 1
                            assert end == start - len(sequence)
                # Record the end
                ends[index] = end
                sequences[index] += sequence
                aligned_sequences[index] += aligned_sequence
                if index == 0:
                    column += len(aligned_sequence)
                else:
                    assert column == len(aligned_sequences[index])
                index += 1
        aligned_sequences = [
            aligned_sequence.encode() for aligned_sequence in aligned_sequences
        ]
        sequences, coordinates = Alignment.parse_printed_alignment(aligned_sequences)
        records = []
        n = len(sequences)
        for i in range(n):
            start = starts[i]
            end = ends[i]
            data = sequences[i]
            if start == 0:
                sequence = Seq(data)
            elif start < end:
                coordinates[i, :] += start
                # create a partially defined sequence
                sequence = Seq({start: data}, length=end)
            else:
                start, end = end, start
                coordinates[i, :] = end - coordinates[i, :]
                # create a partially defined sequence
                sequence = Seq({0: data}, length=end).reverse_complement()
            record = SeqRecord(sequence, identifiers[i])
            records.append(record)
        alignment = Alignment(records, coordinates)
        if annotations:
            alignment.annotations = annotations
        if consensus:
            alignment.column_annotations = {"emboss_consensus": consensus}
        return alignment
